from fmtr.tools import version_tools as version

__version__ = version.read()

import fmtr.tools.config_tools as config
import fmtr.tools.dataclass_tools as dataclass
import fmtr.tools.datatype_tools as datatype
import fmtr.tools.environment_tools as env
import fmtr.tools.environment_tools as environment
import fmtr.tools.function_tools as function
import fmtr.tools.hash_tools as hash
import fmtr.tools.import_tools as import_
import fmtr.tools.iterator_tools as iterator
import fmtr.tools.json_tools as json
import fmtr.tools.path_tools as path
import fmtr.tools.platform_tools as platform
import fmtr.tools.random_tools as random
import fmtr.tools.string_tools as string
import fmtr.tools.name_tools as name
import fmtr.tools.logging_tools as logging
from fmtr.tools.logging_tools import logger

from fmtr.tools.import_tools import MissingExtraMockModule
from fmtr.tools.path_tools import Path

try:
    from fmtr.tools import augmentation_tools as augmentation
except ImportError as exception:
    augmentation = MissingExtraMockModule('augmentation', exception)

try:
    from fmtr.tools import yaml_tools as yaml
except ImportError as exception:
    yaml = MissingExtraMockModule('yaml', exception)

try:
    from fmtr.tools import docker_tools as docker
    from fmtr.tools.docker_tools import Container
except ImportError as exception:
    docker = Container = MissingExtraMockModule('docker.api', exception)

try:
    from fmtr.tools import parallel_tools as parallel
except ImportError as exception:
    parallel = MissingExtraMockModule('parallel', exception)

try:
    from fmtr.tools import profiling_tools as profiling
    from fmtr.tools.profiling_tools import Timer
except ImportError as exception:
    profiling = Timer = MissingExtraMockModule('profiling', exception)

try:
    import fmtr.tools.process_tools as process
    from fmtr.tools.process_tools import ContextProcess
except ImportError as exception:
    process = ContextProcess = MissingExtraMockModule('process', exception)

try:
    from fmtr.tools import tokenization_tools as tokenization
except ImportError as exception:
    tokenization = MissingExtraMockModule('tokenization', exception)

try:
    from fmtr.tools import unicode_tools as unicode
except ImportError as exception:
    unicode = MissingExtraMockModule('unicode', exception)

try:
    from fmtr.tools import netrc_tools as netrc
except ImportError as exception:
    netrc = MissingExtraMockModule('netrc', exception)

try:
    from fmtr.tools import spaces_tools as spaces
except ImportError as exception:
    spaces = MissingExtraMockModule('spaces', exception)

try:
    from fmtr.tools import hfh_tools as hfh
except ImportError as exception:
    hfh = MissingExtraMockModule('hfh', exception)

try:
    from fmtr.tools import merging_tools as merging
    from fmtr.tools.merging_tools import merge
except ImportError as exception:
    merging = merge = MissingExtraMockModule('merging', exception)

try:
    from fmtr.tools import api_tools as api
except ImportError as exception:
    api = MissingExtraMockModule('api', exception)

try:
    from fmtr.tools import ai_tools as ai
except ImportError as exception:
    ai = MissingExtraMockModule('ai', exception)

try:
    from fmtr.tools import data_modelling_tools as dm
except ImportError as exception:
    dm = MissingExtraMockModule('dm', exception)

try:
    from fmtr.tools import json_fix_tools as json_fix
except ImportError as exception:
    json_fix = MissingExtraMockModule('json_fix', exception)

try:
    from fmtr.tools import semantic_tools as semantic
except ImportError as exception:
    semantic = MissingExtraMockModule('semantic', exception)

try:
    from fmtr.tools import metric_tools as metric
except ImportError as exception:
    metric = MissingExtraMockModule('metric', exception)

try:
    from fmtr.tools import html_tools as html
except ImportError as exception:
    html = MissingExtraMockModule('html', exception)

try:
    from fmtr.tools import interface_tools as interface
except ImportError as exception:
    interface = MissingExtraMockModule('interface', exception)

try:
    from fmtr.tools import openai_tools as openai
except ImportError as exception:
    openai = MissingExtraMockModule('openai', exception)

try:
    from fmtr.tools import google_api_tools as google_api
except ImportError as exception:
    google_api = MissingExtraMockModule('google.api', exception)


__all__ = [
    'config',
    'dataclass',
    'datatype',
    'environment',
    'env',
    'function',
    'hash',
    'import_',
    'iterator',
    'json',
    'path',
    'Path',
    'platform',
    'process',
    'ContextProcess',
    'random',
    'string',
    'logging',
    'logger',
    'augmentation',
    'docker',
    'Container',
    'parallel',
    'profiling',
    'Timer',
    'tokenization',
    'unicode',
    'version'
]



import contextlib
import docker


@contextlib.contextmanager
def Container(image, ports=None, name=None, **kwargs):
    """

    Run a Docker container in a context manager

    """
    client = docker.from_env()

    try:
        container = client.containers.get(name)
        container.stop()
        container.remove()
    except docker.errors.NotFound:
        pass

    ports = {f'{port}/tcp': port for port in ports}
    container = client.containers.run(image, ports=ports, detach=True, name=name, **kwargs)

    try:
        yield container
    finally:
        container.stop()
        container.remove()





"""

Config tools for downstream projects

"""
from dataclasses import dataclass, fields, Field as DataclassField

from typing import List, Type


class ConfigClass:
    """

    Base class for all config classes.

    """

    @classmethod
    def process_field(cls, field):
        """

        Post-process field

        """

    @classmethod
    def process_fields(cls):
        """

        Post-process fields

        """
        for field in cls.get_fields():
            cls.process_field(field)

    @classmethod
    def get_fields(cls) -> List[DataclassField]:
        """

        Return fields

        """
        return fields(cls)

    def __init_subclass__(cls, **kwargs):
        """

        Decorate subclasses as dataclasses

        """
        return dataclass(cls)


Field = Type



from deepmerge import always_merger
from functools import reduce


def merge(*data):
    """

    Merges multiple data dictionaries/objects recursively.

    """
    merged = reduce(always_merger.merge, data)
    return merged



from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

from fmtr.tools.logging_tools import logger


class Authenticator:
    """

    Authenticate via link or saved token

    """

    PATH = None
    SCOPES = []
    SERVICE = None
    VERSION = None

    @classmethod
    def auth(cls):
        msg = f'Doing auth for service {cls.SERVICE} ({cls.VERSION})...'
        logger.info(msg)

        PATH_CREDS = cls.PATH / 'credentials.json'
        PATH_TOKEN = cls.PATH / 'token.json'

        if PATH_TOKEN.exists():
            data_token = PATH_TOKEN.read_json()
            credentials = Credentials.from_authorized_user_info(data_token, cls.SCOPES)
        else:
            flow = InstalledAppFlow.from_client_secrets_file(PATH_CREDS, cls.SCOPES)
            credentials = flow.run_local_server(open_browser=False)
            PATH_TOKEN.write_text(credentials.to_json())
        service = build(cls.SERVICE, cls.VERSION, credentials=credentials)
        return service



import re
from collections import namedtuple
from string import Formatter
from typing import List

ELLIPSIS = '…'
formatter = Formatter()

Segment = namedtuple('Segment', ['literal_text', 'field_name', 'format_spec', 'conversion'])


def parse_string(string: str) -> List[Segment]:
    """

    Return structured version of a string with formatting slots.

    """
    parsed = [Segment(*args) for args in formatter.parse(string)]
    return parsed


def is_format_string(string: str) -> bool:
    """

    Does the string contains string formatting slots (i.e. {})?

    """
    try:
        parsed = parse_string(string)
    except ValueError:
        return False
    if all(datum.field_name is None for datum in parsed):
        return False
    else:
        return True


def get_var_name(string: str) -> str:
    """

    Get the name of a variable from a (resolved) f-string `{a=}`

    """
    name, value = string.split('=', maxsplit=1)
    return name


def format_data(value, **kwargs):
    """

    Format a complex object

    """
    if isinstance(value, str):
        return value.format(**kwargs)
    elif isinstance(value, dict):
        return {format_data(k, **kwargs): format_data(v, **kwargs) for k, v in value.items()}
    elif isinstance(value, list):
        return [format_data(item, **kwargs) for item in value]
    else:
        return value


WHITESPACE = re.compile('[\s\-_]+')


def sanitize(*strings, sep: str = '-') -> str:
    """

    Replace spaces with URL- and ID-friendly characters, etc.

    """

    strings = [string for string in strings if string]
    string = ' '.join(strings)
    strings = [c.lower() for c in string if c.isalnum() or c in {' '}]
    string = ''.join(strings)
    string = WHITESPACE.sub(sep, string).strip()

    return string


def truncate_mid(text, length=None, sep=ELLIPSIS):
    """

    Truncate a string to `length` characters in the middle

    """
    if len(text) <= length or not length:
        return text
    half_length = (length - 3) // 2
    return text[:half_length] + sep + text[-half_length:]


def flatten(raw):
    """

    Flatten a multiline string to a single line

    """
    lines = raw.splitlines()
    text = ' '.join(lines)
    text = text.strip()
    return text


class Mask:
    """

    Allows partial-like f-strings

    """

    def __init__(self, mask: str):
        self.mask = mask
        self.kwargs = {}
        self.args = []

    def format(self, *args, **kwargs):
        """

        If the string is complete, return it, else store field values

        """
        self.args += list(args)
        self.kwargs.update(kwargs)
        try:
            text = self.mask.format(*args, **self.kwargs)
            return text
        except (KeyError, IndexError):
            return self



import torch
from datetime import datetime
from peft import PeftConfig, PeftModel
from pydantic import Field
from statistics import mean
from transformers import AutoModelForCausalLM, AutoTokenizer
from typing import List

from fmtr.tools import logger, data_modelling_tools
from fmtr.tools.hfh_tools import get_hf_cache_path

CPU = 'cpu'
GPU = 'cuda'


class DynamicBatcher:
    """

    Helper to simplify dynamic batching.

    """

    def __init__(self, prompts, threshold_stable_reset=None, factor_reduction=None, size_max=100):

        self.prompts = prompts
        self.size_total = len(self)
        self.size_max = size_max or self.size_total
        self.threshold_stable_reset = threshold_stable_reset or 50
        self.factor_reduction = factor_reduction or 0.9
        self.size = self.count_stable = None
        self.started = datetime.now()

        self.reset()

    def reset(self):
        """

        Reset inferred, stable batch size. Useful if outlier outputs force the batch size below what's optimal

        """
        self.size = min([len(self), self.size_max])
        self.count_stable = 0

    def batch_complete(self):
        """

        When a batch completes, remove it. If the number of stable batches reaches the threshold, trigger reset.

        """

        self.prompts = self.prompts[self.size:]

        self.count_stable += 1

        percent = 100 - ((len(self) / self.size_total) * 100)

        msg = (
            f"Marking batch (size {self.size}) complete. {len(self)} of {self.size_total} prompts remaining. "
        )
        logger.info(msg)
        msg = (
            f"Stable for {self.count_stable}/{self.threshold_stable_reset} batch(es). "
            f"{percent:.2f}% complete. "
            f"Elapsed: {datetime.now() - self.started}. "
            f"Estimated: {self.calculate_eta()}."
        )
        logger.info(msg)

        if self.count_stable >= self.threshold_stable_reset:
            msg = (f"Stable count reached threshold of {self.threshold_stable_reset}. Resetting batch size.")
            logger.info(msg)
            self.reset()

    def __len__(self):
        """

        Length is number of remaining prompts

        """
        return len(self.prompts)

    def get(self):
        """

        Fetch a batch of current size

        """
        logger.info(f"Processing batch of {self.size} prompts...")
        return self.prompts[:self.size]

    def reduce(self):
        """

        If OOM occurs, reduce batch size by specified factor, or by at least 1.

        """

        self.count_stable = 0

        size_new = round(self.size * self.factor_reduction)

        if size_new == self.size:
            self.size -= 1
            logger.info(f"Batch size reduced to {self.size} prompts")
        else:
            self.size = size_new
            logger.info(f"Batch size reduced by factor {self.factor_reduction} to {self.size} prompts")

        if self.size == 0:
            if len(self) < self.size_total:
                msg = f"Batch size 1 caused OOM, despite previous batches succeeding. Will retry. Try freeing resources."
                logger.warning(msg)
                self.size = 1
            else:
                raise ValueError('Size of first batch reached 0. Prompt(s) are likely extremely long.')

    def calculate_eta(self):
        """

        Calculate bulk-job ETA.

        """
        time_spent = datetime.now() - self.started

        completed = self.size_total - len(self)
        if completed <= 0:
            return "Unknown"

        average_time_per_task = time_spent / completed
        remaining_time = average_time_per_task * len(self)
        eta = datetime.now() + remaining_time

        return eta


class BulkInferenceManager:
    """

    Perform bulk LLM inference using the specified configuration and dynamic batching.

    """

    LOCAL_ONLY = False
    PRECISION_FLOAT = torch.float16

    REPO_ID = 'mistralai/Mistral-7B-Instruct-v0.3'
    REPO_ID_ADAPTER = None
    REPO_TAG_ADAPTER = 'main'

    DEVICE_ACTIVE = GPU
    DEVICE_INACTIVE = CPU

    BATCH_STABLE_RESET = None
    BATCH_FACTOR_REDUCTION = None
    BATCH_SIZE_MAX = None

    TOOLS = None

    def __init__(self):
        """

        Load a base model plus optional adapter in the specified precision, then deactivate until first use.

        """

        args = dict(local_files_only=self.LOCAL_ONLY, torch_dtype=self.PRECISION_FLOAT)

        logger.info(f"Loading base model from {self.REPO_ID}")
        base_model = AutoModelForCausalLM.from_pretrained(self.REPO_ID, **args)
        self.tokenizer = AutoTokenizer.from_pretrained(self.REPO_ID, **args)
        self.tokenizer.pad_token = self.tokenizer.eos_token

        if self.REPO_ID_ADAPTER:
            logger.info(f"Loading adapter model from {self.REPO_ID_ADAPTER} [{self.REPO_TAG_ADAPTER}]")
            if self.LOCAL_ONLY:
                path_adapter = get_hf_cache_path(self.REPO_ID_ADAPTER, tag=self.REPO_TAG_ADAPTER)
            else:
                path_adapter = self.REPO_ID_ADAPTER
            self.adapter_config = PeftConfig.from_pretrained(path_adapter, revision=self.REPO_TAG_ADAPTER, **args)
            self.model = PeftModel.from_pretrained(base_model, path_adapter, revision=self.REPO_TAG_ADAPTER, **args)
            self.model_adapter = self.model
        else:
            self.model = base_model
            self.adapter_config = self.model_adapter = None

        self.deactivate()

    def activate(self):
        """

        Move the model to the specified active device.

        """
        if self.model.device != self.DEVICE_ACTIVE:
            logger.info(f'Activating model {self.REPO_ID_ADAPTER or self.REPO_ID} to device {self.DEVICE_ACTIVE}')
            self.model = self.model.to(self.DEVICE_ACTIVE)

    def deactivate(self):
        """

        Move the model to the specified active inactive.

        """
        if self.model.device != self.DEVICE_INACTIVE:
            logger.info(f'Deactivating model {self.REPO_ID_ADAPTER or self.REPO_ID} to device {self.DEVICE_INACTIVE}')
            self.model = self.model.to(self.DEVICE_INACTIVE)

    def encode(self, prompts: List[str]):
        """

        Encode/tokenize a list of text prompts to a batch of tokens, including appropriate templates, etc.

        """

        messages = [[{"role": "user", "content": prompt}] for prompt in prompts]

        ids_input = self.tokenizer.apply_chat_template(
            messages,
            return_tensors="pt",
            tools=self.TOOLS,
            padding=True,
            return_attention_mask=True,
            return_dict=True
        )

        return ids_input

    def generate(self, prompts, **params):
        """

        Generate outputs for all batches, using dynamic batching, backoff in case off OOM errors, etc.

        """
        logger.info(f'Starting generation...')
        logger.debug(f'Generation parameters: {params}')

        batcher = DynamicBatcher(
            prompts,
            factor_reduction=self.BATCH_FACTOR_REDUCTION,
            threshold_stable_reset=self.BATCH_STABLE_RESET,
            size_max=self.BATCH_SIZE_MAX
        )

        self.activate()

        while len(batcher):
            try:

                prompts = batcher.get()

                batch_encoding = self.encode(prompts).to(self.model.device)

                ids_input, ids_attention = batch_encoding.data['input_ids'], batch_encoding.data['attention_mask']

                ids_output = self.model.generate(
                    pad_token_id=self.tokenizer.eos_token_id,
                    input_ids=ids_input,
                    attention_mask=ids_attention,
                    **params
                )
                ids_output = ids_output.to(self.DEVICE_INACTIVE)
                batcher.batch_complete()

                ids_output = self.remove_prompt_ids(ids_input, ids_output)

                yield ids_output

            except RuntimeError as exception:
                if "CUDA out of memory" in str(exception):
                    logger.warning(f"Ran out of memory. Reducing batch size: {repr(exception)}")
                    batcher.reduce()
                else:
                    raise

        self.deactivate()
        logger.info(f'Generation complete.')

    def remove_prompt_ids(self, ids_input, ids_output):
        """

        Outputs contain the prompt & the prompt is left-padded, so we can remove it by removing the width of the inputs

        """
        width = ids_input.shape[-1]
        ids_output = ids_output[:, width:]
        return ids_output

    def decode(self, ids_output):
        """

        Decode outputs to text

        """
        texts = self.tokenizer.batch_decode(ids_output, skip_special_tokens=True)
        return texts

    def get_outputs(self, prompts: List[str], **params):
        """

        Generate a batch of outputs from a batch of prompts

        """

        params = params or dict(do_sample=False)

        for ids_output in self.generate(prompts, **params):
            texts = self.decode(ids_output)

            lengths = [len(text) // 5 for text in texts]
            msg = f'Text statistics: {min(lengths)=} {max(lengths)=} {mean(lengths)=}.'
            logger.info(msg)

            yield from texts

    def get_output(self, prompt, **kwargs):
        """

        Get a singleton output

        """
        outputs = self.get_outputs([prompt], **kwargs)
        output = next(iter(outputs))
        return output


class ToolCall(data_modelling_tools.Base):
    """

    Tool call data

    """
    name: str = Field(
        ...,
        description="The function name"
    )
    arguments: dict = Field(
        ...,
        description="The function arguments"
    )

    def apply(self, functions):
        """

        Apply the specified functions to their arguments

        """
        functions = {function.__name__: function for function in functions}
        function = functions[self.name]
        obj = function(**self.arguments)
        return obj


class ToolsCall(data_modelling_tools.Root):
    """

    Tool calls data

    """
    root: List[ToolCall]

    def apply(self, functions):
        objs = [child.apply(functions) for child in self.root]
        return objs

def tst():
    """

    Test with a large number of small input/outputs.

    """
    mask = 'Write out the following number as words: {}. Just the text please, no explanation or alternatives'
    prompts = [mask.format(i) for i in range(10_000)]
    manager = BulkInferenceManager()
    texts = manager.get_outputs(prompts, max_new_tokens=20, do_sample=True, temperature=1.2, top_p=0.5, top_k=50)

    data = {}
    for prompt, text in zip(prompts, texts):
        data[prompt] = text
    return data


def tst_tool():
    """

    Test Tool usage

    """

    def get_current_weather(location: str, format: str):
        """
        Get the current weather

        Args:
            location: The city and state, e.g. San Francisco, CA
            format: The temperature unit to use. Infer this from the users location. (choices: ["celsius", "fahrenheit"])
        """
        return "It's 25 degrees and sunny!"

    class BulkInferenceManagerTools(BulkInferenceManager):
        TOOLS = [get_current_weather]

    prompt = "What's the weather like in Paris?"
    prompts = [prompt]
    manager = BulkInferenceManagerTools()
    gen = manager.get_outputs(prompts, max_new_tokens=200, do_sample=True, temperature=1.2, top_p=0.5, top_k=50)
    texts = list(gen)

    for text in texts:
        objs = ToolsCall.from_json(text).apply(BulkInferenceManagerTools.TOOLS)
        obj = objs[0]
        print(obj)





if __name__ == '__main__':
    texts = tst_tool()
    texts
